import java.io.*;
import java.util.*;
import org.w3c.dom.*;
import org.xml.sax.*;
import com.ibm.xml.parsers.*;

/**
 * XMLElementSearch class is basically used to Search Elements in an XML Document
 */
public class XMLElementSearch{
	/**
	 * This variable is used to specify the Search is pattern is Node
	 */
	public static final int ELEMENT_SEARCH = 1;

	/**
	 * This variable is used to specify the Search is pattern is Text
	 */
	public static final int TEXT_SEARCH = 2;

	private Vector store;
	private String elementName;
	private int searchPattern = XMLElementSearch.TEXT_SEARCH;
	private DOMParser parser;
	private boolean domCreated = false;
	private String xmlString;
	private InputSource source;

	/**
	 * Constructor which accepts the XML Document
	 *
	 * @param xmlString The XML Document
 	 * @exception XMLElementSearchException when could not parse the document
	 */
	public XMLElementSearch(String xmlString) throws XMLElementSearchException{
		setXML(xmlString);
	}

	/**
	 * Constructor which accepts the XML Document
	 *
	 * @param reader The stream from where the XML Document can be read
 	 * @exception java.lang.Exception when could not parse the document
	 */
	public XMLElementSearch(Reader reader) throws XMLElementSearchException{
		setXML(reader);
	}

	/**
	 * Constructor which accepts the XML Document
	 *
	 * @param inputStream The stream from where the XML Document can be read
 	 * @exception java.lang.Exception when could not parse the document
	 */
	public XMLElementSearch(InputStream inputStream) throws XMLElementSearchException{
		setXML(inputStream);
	}

	// initializing the DOM
	private void init() throws XMLElementSearchException{
		store = new Vector();
		if( domCreated ){
			parser.reset();
		}else{
			parser = new DOMParser();
			domCreated = true;
		}
		try{
			parser.parse(source);
		}catch(IOException e){
			throw new XMLElementSearchException("I/O Exception "+e.getMessage());
		}catch(SAXException sax){
			throw new XMLElementSearchException("Unable to parse the Document "+sax.getMessage());
		}
	}

	/**
	 * Method for setting the XML Document
	 *
	 * @param xmlString The XML Document
 	 * @exception XMLElementSearchException when could not load the document
	 */
	public void setXML(String xmlString) throws XMLElementSearchException{
		this.xmlString = xmlString;
		StringReader read = new StringReader(xmlString);
		source = new InputSource(read);
		init();
	}

	/**
	 * Method for setting the XML Document
	 *
	 * @param reader The stream from where the XML Document can be read
 	 * @exception XMLElementSearchException when could not load the document
	 */
	public void setXML(Reader reader) throws XMLElementSearchException{
		source = new InputSource(reader);
		init();
	}

	/**
	 * Method for setting the XML Document
	 *
	 * @param inputStream The stream from where the XML Document can be read
 	 * @exception XMLElementSearchException when could not load the document
	 */
	public void setXML(InputStream inputStream) throws XMLElementSearchException{
		source = new InputSource(inputStream);
		init();
	}

	/**
	 * Method for searching for a specifing element in the XML Document.
	 * If the searchPattern is XMLElementSearch.TEXT_SEARCH then from the Whole
	 * XML Document if the Element occurs it returns the Value of the Element
	 * for eg: in <NAME>Levi</NAME> if you are searching for elementName=NAME
	 * then the search will return a Vector of the Values. in this case calling
	 * Vector.elementAt(0) will return the value 'Levi'. If the elementName
	 * occurs multiple times in the same document search will return Vector of
	 * values.<br>
	 * If searchPattern=XMLElementSearch.ELEMENT_SEARCH then the search will
	 * returns a Vector of <b>Node</b> objects
	 *
	 * @param elementName The Element Name to be searched
	 * @param searchPattern search Pattern can be XMLElementSearch.ELEMENT_SEARCH
	 *                      or XMLElementSearch.TEXT_SEARCH
     * @return The Vector object which has either the String Values or Node Objects
     *         depending upon the searchPattern
	 */
	public Vector search(String elementName,int searchPattern){
		store = new Vector();
		this.elementName = elementName;
		this.searchPattern = searchPattern;
		Document doc = parser.getDocument();
		if (doc != null){
			parseNode(doc);
		}
		return store;
	}

	/**
	 * Method for searching for a specifing element in the XML Document.
	 *
	 * @param node The node under which you want search for an Element
     * @return The Vector object which has either the String Values or Node Objects
     *         depending upon the searchPattern
	 */
	public Vector search(Node node){
		store = new Vector();
		parseNode(node);
		return store;
	}

	/**
	 * Method for searching for a specifing element in the XML Document.
	 * If the searchPattern is XMLElementSearch.TEXT_SEARCH then from the Whole
	 * XML Document if the Element occurs it returns the Value of the Element
	 * for eg: in <NAME>Levi</NAME> if you are searching for elementName=NAME
	 * then the search will return a Vector of the Values. in this case calling
	 * Vector.elementAt(0) will return the value 'Levi'. If the elementName
	 * occurs multiple times in the same document search will return Vector of
	 * values.<br>
	 * If searchPattern=XMLElementSearch.ELEMENT_SEARCH then the search will
	 * returns a Vector of <b>Node</b> objects
	 *
	 * @param node The node under which you want search for an Element
	 * @param elementName The Element Name to be searched
	 * @param searchPattern search Pattern can be XMLElementSearch.ELEMENT_SEARCH
	 *                      or XMLElementSearch.TEXT_SEARCH
     * @return The Vector object which has either the String Values or Node Objects
     *         depending upon the searchPattern
	 */
	public Vector search(Node node,String elementName,int searchPattern){
		store = new Vector();
		this.elementName = elementName;
		this.searchPattern = searchPattern;
		parseNode(node);
		return store;
	}

	/**
	 * Method for searching for a specifing element in the XML Document and retrieve its value.
	 * The search pattern is now specified by the attribute name, values stored in
	 * Hashtable. If no attributes then it will still still return the Element's
	 * Value if any found in the Document.
	 *
	 * @param elementName The Element Name to be searched
	 * @param hash The hash table which has the Attribute Names and Values
     * @return The Vector object which has the String Value(s)
	 */
	public Vector getElementValueByAttributes(String elementName,Hashtable hash){
		this.elementName = elementName;
		this.searchPattern = XMLElementSearch.ELEMENT_SEARCH;
		store = new Vector();
		parseNode(parser.getDocument());
		if( store.size() == 0 ){
			return null;
		}else if( hash == null ){
			return search(elementName,XMLElementSearch.TEXT_SEARCH);
		}

		Vector values = new Vector(2,1);
		Vector holder = store;

		for( int i=0 ; i < holder.size() ; i++ ){
			Node node = (Node)holder.elementAt(i);

			Hashtable source = getAttributes(node);
			Enumeration enum = hash.keys();
			boolean allKeysMatched = true;

			while( enum.hasMoreElements() ){
				String key = (String)enum.nextElement();
				String val1 = (String)hash.get(key);
				Object dupVal = source.get(key);
				String val2 = null;
				if( dupVal != null && dupVal instanceof String ){
					val2 = (String)dupVal;
				}
				if( val2 == null || !val2.trim().equals(val1) ){
					allKeysMatched = false;
					break;
				}
			}
			if( allKeysMatched ){
				this.elementName = elementName;
				this.searchPattern = XMLElementSearch.TEXT_SEARCH;
				Vector temp = search(node);
				for( int j=0; j<temp.size() ; j++ ){
					values.addElement(temp.elementAt(j));
				}
			}
		}
		if( values.size() == 0 ){
			return null;
		}
		return values;
	}

	/**
	 * Method for setting the Element Name for search
	 *
	 * @param elementName The Element Name to be searched
	 */
	public void setElementName(String elementName){
		this.elementName = elementName;
	}

	/**
	 * Method for setting the search pattern. You can set either
	 * XMLElementSearch.TEXT_SEARCH OR XMLElementSearch.ELEMENT_SEARCH
	 *
	 * @param searchPattern The search pattern
	 */
	public void setSearchPattern(int searchPattern){
		this.searchPattern = searchPattern;
	}

	/**
	 * Method for setting the search pattern. You can set either
	 * XMLElementSearch.TEXT_SEARCH OR XMLElementSearch.ELEMENT_SEARCH
	 *
	 * @param node The Node object for which you want to retrieve the Attributes
	 * @return The attribute names and values for a specific Element Node
	 */
	public Hashtable getAttributes(Node node){
		NamedNodeMap attrs = node.getAttributes();
		if( attrs == null || attrs.getLength() == 0 ){
			return null;
		}
		Hashtable hash = new Hashtable();
		for (int i = 0; i < attrs.getLength(); i++){
			Node attr = attrs.item(i);
			hash.put(attr.getNodeName(),attr.getNodeValue());
		}
		return hash;
	}

	/**
	 * Method for searching for a specified Node under a give node.
	 * This search is a vertical search not a tree search.
	 *
	 * @param node The Node object under which you want to search for a specific Node(Element)
	 * @param nodeName the node Name for which you want to search
	 * @return The Vector of node(s) found under the specified node
	 */
	public Vector searchForNodeUnderElement(Node node,String nodeName){
		Vector nodes = new Vector(5,1);
		Node next = node.getFirstChild();
		while( next != null ){
			if( next.getNodeType() == Node.ELEMENT_NODE && next.getNodeName().equals(nodeName) ){
				nodes.addElement(next);
			}
			next = next.getNextSibling();
		}
		if( nodes.size() == 0 ){
			return null;
		}
		return nodes;
	}

	/**
	 * For internal purpose
	 *
	 * @param node the Node you want to parse and search for a specific element for a given
	 *             pattern
	 */
	public void parseNode(Node node)
	{
		int type = node.getNodeType();
		switch (type){
			case Node.DOCUMENT_NODE:{
				parseNode(((Document)node).getDocumentElement());
				break;
			}
			case Node.ELEMENT_NODE:{
				if( searchPattern == XMLElementSearch.ELEMENT_SEARCH){
					if( elementName.equals(node.getNodeName()) ){
						store.addElement(node);
					}
				}
				NodeList children = node.getChildNodes();
				if (children != null){
					int len = children.getLength();
					for (int i = 0; i < len; i++)
					parseNode(children.item(i));
				}
				break;
			}

			case Node.ENTITY_REFERENCE_NODE:{
				break;
			}

			case Node.CDATA_SECTION_NODE:{
				break;
			}

			case Node.TEXT_NODE:{
				if( searchPattern == XMLElementSearch.TEXT_SEARCH){
					Node parent = node.getParentNode();
					if( elementName.equals(parent.getNodeName()) ){
						store.addElement(node.getNodeValue());
					}
				}
				break;
			}

			case Node.PROCESSING_INSTRUCTION_NODE:{
				break;
			}
		}
	}
}